package com.alterdekim.flash.decompiler.mapper;

import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.tuple.Pair;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.stream.Collectors;

@Slf4j
public class FlashMapper {

    public static <T> Map<String, Object> objToMap(List<T> l) {
        Map<String, Object> m = new HashMap<>();
        for( T o : l ) {
            try {
                Map<String, Object> m1 = new HashMap<>();
                String s = getId(o).toString();
                List<Field> fields = Arrays.stream(o.getClass().getDeclaredFields())
                        .filter(f -> f.isAnnotationPresent(FlashField.class))
                        .collect(Collectors.toList());
                for (Field f : fields) {
                    String key = f.getAnnotation(FlashField.class).name();
                    Object val = f.get(o);
                    m1.put(key, val);
                }
                m.put(s, m1);
            } catch (IllegalAccessException e) {
                log.error("FlashMapper valueGet error: {}", e.getMessage());
            }
        }
        return m;
    }

    private static <T> Long getId(T o) throws IllegalAccessException {
        Optional<Field> idField = Arrays.stream(o.getClass().getDeclaredFields())
                .filter(f -> f.isAnnotationPresent(FlashValue.class) &&
                        f.getAnnotation(FlashValue.class).type() == FlashValueType.Id)
                .findFirst();
        if( idField.isEmpty() ) return 0L;
        Field f = idField.get();
        f.setAccessible(true);
        return (Long) f.get(o);
    }

    public static <T> T mapToObj(Map<String, Object> m, Class<T> obj) {
        try {
            T result = obj.getDeclaredConstructor().newInstance();
            Arrays.stream(obj.getDeclaredFields())
                    .filter(f -> m.containsKey(f.getName()) || (f.isAnnotationPresent(FlashField.class) && m.containsKey(f.getAnnotation(FlashField.class).name())))
                    .map(f -> Pair.of(f, f.isAnnotationPresent(FlashField.class) ? f.getAnnotation(FlashField.class).name() : f.getName()))
                    .forEach(p -> {
                        try {
                            Field f = p.getLeft();
                            String name = p.getRight();
                            f.setAccessible(true);
                            if( f.isAnnotationPresent(FlashField.class) && f.getAnnotation(FlashField.class).useCustomConverter()) {
                                f.set(result, f.getAnnotation(FlashField.class).converter().getDeclaredConstructor().newInstance().convert(m.get(name)));
                                return;
                            }
                            f.set(result, m.get(name));
                        } catch (IllegalAccessException | NoSuchMethodException | InstantiationException |
                                 InvocationTargetException e) {
                            log.error("FlashMapper valueSet error: {}", e.getMessage());
                        }
                    });
            return result;
        } catch (Exception e) {
            log.error("FlashMapper error: {}", e.getMessage());
        }
        return null;
    }

    public static void mapToObj(Map<String, Object> o, List<OutputObject> objs) {
        for( OutputObject object : objs ) {
            String entryName = object.getObjectClass().getAnnotation(FlashClass.class).name();
            Map<String, Object> m = (Map<String, Object>) o.get(entryName);
            if( object.getIsSimple() ) {
                m.keySet().forEach(k -> {
                    try {
                        Object entry = object.getObjectClass().getDeclaredConstructor().newInstance();

                        Arrays.stream(object.getObjectClass().getDeclaredFields())
                                        .filter(f -> f.isAnnotationPresent(FlashValue.class))
                                        .forEach(f -> {
                                            try {
                                                f.setAccessible(true);
                                                switch (f.getAnnotation(FlashValue.class).type()) {
                                                    case Id -> f.set(entry, Long.parseLong(k));
                                                    case Value -> f.set(entry, m.get(k));
                                                }
                                            } catch (IllegalAccessException e) {
                                                log.error("FlashMapper valueSet error: {}", e);
                                            }
                                        });
                        object.getCallback().onOutputObjectReady(entry);
                    } catch (InstantiationException | IllegalAccessException | InvocationTargetException |
                             NoSuchMethodException e) {
                        log.error("FlashMapper class error: {}", e.getMessage());
                    }
                });
                continue;
            }

            m.keySet().forEach(k -> {
                Long id = Long.parseLong(k);
                Map<String, Object> obj = (Map<String, Object>) m.get(k);
                Object entity = FlashMapper.mapToObj(obj, object.getObjectClass());
                Optional<Field> f1 = Arrays.stream(object.getObjectClass().getDeclaredFields())
                        .filter(f -> f.isAnnotationPresent(FlashValue.class) && f.getAnnotation(FlashValue.class).type() == FlashValueType.Id)
                        .findFirst();
                if( f1.isEmpty() ) {
                    log.error("FlashMapper valueSet error; can't find Id field for object.");
                    return;
                }
                f1.get().setAccessible(true);
                try {
                    f1.get().set(entity, id);
                } catch (IllegalAccessException e) {
                    log.error("FlashMapper valueSet error: {}", e);
                    return;
                }
                object.getCallback().onOutputObjectReady(entity);
            });
        }
    }
}
